﻿/*Copyright 2024 Wojciech Krzysztof Fiałkiewicz

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using Microsoft.CSharp;
using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;

namespace FunctionsCompiler
{
    public class Compiler
    {
        public static Functions.Function2DT GetCompiledFunction2DT(string text)
        {
            object m_RuntimeCreatedObject;
            CompilerResults m_CompilerResults;

            if (!CheckIfSafe(text))
            {
                return null;
            }

            text = AddCallingFunctions(text);

            string code = @"using Functions;
                            using System.Collections.Generic;
     

                            namespace RuntimeGenerated
                            {
                                public class CalculateFunction : Function2DT
                                {
                                     public double Calculate(ref List<double> prevs, ref List<FunctionsAdapter.FUN> funcs, double x, double z, double t)
                                     {
                                         bool connect;

                                         " + text + @"; 
                                     }
                                 }    
                            }";

            Dictionary<string, string> providerOptions = new Dictionary<string, string>
                {
                    {"CompilerVersion", "v4.0"}
                };
            CSharpCodeProvider provider = new CSharpCodeProvider(providerOptions);

            CompilerParameters compilerParams = new CompilerParameters
            {
                GenerateInMemory = true,
                GenerateExecutable = false
            };

            compilerParams.ReferencedAssemblies.Add("Functions.dll");

            m_CompilerResults = provider.CompileAssemblyFromSource(compilerParams, code);

            if (m_CompilerResults.Errors.Count != 0)
            {
                return null;
            }

            m_RuntimeCreatedObject = m_CompilerResults.CompiledAssembly.CreateInstance("RuntimeGenerated.CalculateFunction");

            return m_RuntimeCreatedObject as Functions.Function2DT;
        }

        public static Functions.Function3DT GetCompiledFunction3DT(string text)
        {
            object m_RuntimeCreatedObject;
            CompilerResults m_CompilerResults;

            if (!CheckIfSafe(text))
            {
                return null;
            }

            text = AddCallingFunctions(text);

            string code = @"using Functions;
                            using System.Collections.Generic;

                            namespace RuntimeGenerated
                            {
                                public class CalculateFunction : Function3DT
                                {
                                     public double Calculate(ref List<double> prevs, ref List<FunctionsAdapter.FUN> funcs, double x, double z, double y, double t)
                                     {
                                         bool connect;

                                         " + text + @"; 
                                     }
                                 }    
                             }";

            Dictionary<string, string> providerOptions = new Dictionary<string, string>
                {
                    {"CompilerVersion", "v4.0"}
                };
            CSharpCodeProvider provider = new CSharpCodeProvider(providerOptions);

            CompilerParameters compilerParams = new CompilerParameters
            {
                GenerateInMemory = true,
                GenerateExecutable = false
            };

            compilerParams.ReferencedAssemblies.Add("Functions.dll");

            m_CompilerResults = provider.CompileAssemblyFromSource(compilerParams, code);

            if (m_CompilerResults.Errors.Count != 0)
            {
                return null;
            }

            m_RuntimeCreatedObject = m_CompilerResults.CompiledAssembly.CreateInstance("RuntimeGenerated.CalculateFunction");

            return m_RuntimeCreatedObject as Functions.Function3DT;
        }

        private static bool CheckIfSafe(string text)
        {
            foreach(char c in text)
            {
                if (c >= 'a' && c <= 'z')
                    continue;

                if (c >= 'A' && c <= 'Z')
                    continue;

                if (c >= '0' && c <= '9')
                    continue;

                if (c == '(' || c == ')' || c == '.')
                    continue;

                if (c == '+' || c == '-' || c == '*' || c == '/')
                    continue;

                if (c == ' ')
                    continue;

                return false;
            }            
            return true;
        }

        private static string AddCallingFunctions(string text)
        {
            text = text.Replace("csc(", "adapter.Csc(");
            text = text.Replace("sec(", "adapter.Sec(");
            text = text.Replace("ln(", "adapter.Ln(");
            text = text.Replace("tan(", "adapter.Tan(");
            text = text.Replace("floor(", "adapter.Floor(");
            text = text.Replace("ceil(", "adapter.Ceil(");
            text = text.Replace("cot(", "adapter.Cot(");
            text = text.Replace("arcsin(", "adapter.ArcSin(");
            text = text.Replace("arccos(", "adapter.ArcCos(");
            text = text.Replace("sin(", "adapter.Sin(");
            text = text.Replace("cos(", "adapter.Cos(");
            text = text.Replace("exp(", "adapter.Exp(");
            text = text.Replace("sinh(", "adapter.Sinh(");
            text = text.Replace("cosh(", "adapter.Cosh(");

            text = "FunctionsAdapter adapter = new FunctionsAdapter(); adapter.Init(prevs);" + Environment.NewLine + "double val = " + text + ";";
            text = text + Environment.NewLine + "if(val > 10000){val = 10000;}else if(val < -10000){val = -10000;}";
            text = text + Environment.NewLine + "if(double.IsPositiveInfinity(val)){val = 10000;}else if(double.IsNegativeInfinity(val)){val = -10000;}";
            text = text + Environment.NewLine + "prevs = adapter.GetPreviousValues();";
            text = text + Environment.NewLine + "connect = adapter.Connect;";
            text = text + Environment.NewLine + "funcs = adapter.GetCalculatedFunctionsOrder();";
            text = text + Environment.NewLine + "return val;";

            return text;
        }

        public static Functions.Function1DT GetCompiledFunction1DT(string text)
        {
            object m_RuntimeCreatedObject;
            CompilerResults m_CompilerResults;

            if (!CheckIfSafe(text))
            {
                return null;
            }

            text = AddCallingFunctions(text);

            string code = @"using Functions;
                            using System.Collections.Generic;


                           namespace RuntimeGenerated
                           {
                              public class CalculateFunction : Function1DT
                              {
                                   public double Calculate(ref List<double> prevs, double x, double t, out bool connect)
                                   {
                                       List<FunctionsAdapter.FUN> funcs;

                                       " + text + @"; 
                                   }
                               }    
                           }";

            Dictionary<string, string> providerOptions = new Dictionary<string, string>
                {
                    {"CompilerVersion", "v4.0"}
                };
            CSharpCodeProvider provider = new CSharpCodeProvider(providerOptions);

            CompilerParameters compilerParams = new CompilerParameters
            {
                GenerateInMemory = true,
                GenerateExecutable = false
            };

            compilerParams.ReferencedAssemblies.Add("Functions.dll");
            compilerParams.ReferencedAssemblies.Add("System.Collections.dll");

            m_CompilerResults = provider.CompileAssemblyFromSource(compilerParams, code);

            if (m_CompilerResults.Errors.Count != 0)
            {
                return null;
            }

            m_RuntimeCreatedObject = m_CompilerResults.CompiledAssembly.CreateInstance("RuntimeGenerated.CalculateFunction");

            return m_RuntimeCreatedObject as Functions.Function1DT;
        }
    }
}
